body {
  padding-top: 2em;
}
.tp-main h1 {
  text-align: center;
  font-size: 4em;
  margin-bottom: 1em;
}
.tp-main .lead {
  font-size: 2em;
}
form {
  display: inline;
}
form input {
  vertical-align: middle;
}
.row {
  text-align: center;
}
.col-md-3 {
  margin-bottom: 1em;
}
